#include <J3ML/LinearAlgebra/AxisAngle.hpp>
#include <J3ML/LinearAlgebra/Quaternion.hpp>
#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
namespace J3ML::LinearAlgebra {


    AxisAngle::AxisAngle(const Vector3& axis, float angle) : axis(axis), angle(angle) {}


    AxisAngle::AxisAngle(const Quaternion& rhs) {
        float halfAngle = std::acos(rhs.w);
        angle = halfAngle * 2.f;
        float reciprocalSinAngle = 1.f / std::sqrt(1.f - rhs.w*rhs.w);

        axis = { rhs.x*reciprocalSinAngle, rhs.y*reciprocalSinAngle, rhs.z*reciprocalSinAngle };
    }


    Quaternion AxisAngle::ToQuaternion() const { return Quaternion(*this); }

    AxisAngle::AxisAngle(const Matrix3x3 &m) : AxisAngle(Quaternion(m)) { }

    bool AxisAngle::Equals(const AxisAngle &rhs, float epsilon) {
        return this->axis.Equals(rhs.axis, epsilon) && Math::Equal(angle, rhs.angle, epsilon);
    }

    Matrix3x3 AxisAngle::ToMatrix3x3() const {
        return Matrix3x3(*this);
    }

    void AxisAngle::Inverse() {
        angle = -angle;
    }

    AxisAngle AxisAngle::Inverted() const {
        return {axis, -angle};
    }

    AxisAngle AxisAngle::Lerp(const AxisAngle &rhs, float t) {
        auto new_axis = axis.Lerp(rhs.axis, t);
        float new_angle = Math::Lerp(angle, rhs.angle, t);
        return {new_axis, new_angle};
    }

    AxisAngle AxisAngle::Slerp(const AxisAngle &rhs, float t) {
        Quaternion a(*this);
        Quaternion b(rhs);

        Quaternion intermediate = a.Slerp(b, t);

        return AxisAngle(intermediate);
    }

    bool AxisAngle::operator==(const AxisAngle &rhs) {
        return Equals(rhs);
    }

    AxisAngle AxisAngle::Normalized() const {
        AxisAngle copy(*this);
        copy.Normalize();
        return copy;
    }

    bool AxisAngle::IsIdentity() { return Math::Equal(angle, 0); }

    void AxisAngle::Normalize() {
        float axisLength = axis.Length();

        if (axisLength > 0.0f)
            axis /= axisLength;
        else {
            // Handle the case where the axis is a zero vector.
            // You might want to set it to a default axis (e.g., (0,1,0))
            axis = Vector3(0, 1, 0);
            angle = 0;
        }
    }

    Vector3 AxisAngle::Axis() const { return axis;}

    float AxisAngle::Angle() const { return angle;}
}